### ---------- IMPORT DEPENDENCIES ----------
from ._pp import _corr_distance, _neighbors_knn, _neighbors_graph

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** DISTANCE FUNCS ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def corr_distance(adata,
                  use_reduction=True,
                  reduction_slot="X_pca",
                  key_added="corr_dist"):
    """\
    A tool for computing a distance matrix based on pearson correlation.

    Parameters
    ----------
    adata
        An anndata object containing a signature in adata.X
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for computing distance.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    key_added (default: "corr_dist")
        Slot in obsp to store the resulting distance matrix.

    Returns
    -------
    Adds fields to the input adata, such that it contains a distance matrix
    stored in adata.obsp[key_added].
    """
    # returns if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
    return _corr_distance(adata, use_reduction, reduction_slot, key_added)

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** KNN ARRAY FUNC ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def neighbors_knn(adata,
                  max_knn=101,
                  dist_slot="corr_dist",
                  key_added="knn",
                  njobs = 1):
    """\
    A tool for computing a KNN array used to then rapidly generate connectivity
    graphs with acdc.pp.neighbors_graph for clustering.

    Parameters
    ----------
    adata
        An anndata object containing a distance object in adata.obsp.
    max_knn (default: 101)
        The maximum number of k-nearest neighbors (knn) to include in this array.
        acdc.pp.neighbors_graph will only be able to compute KNN graphs with
        knn <= max_knn.
    dist_slot (default: "corr_dist")
        The slot in adata.obsp where the distance object is stored. One way of
        generating this object is with adata.pp.corr_distance.
    key_added (default: "knn")
        Slot in uns to store the resulting knn array.
    njobs (default: 1)
        Paralleization option that allows users to speed up runtime.

    Returns
    -------
    Adds fields to the input adata, such that it contains a knn array stored in
    adata.uns[key_added].
    """
    # returns if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
    return _neighbors_knn(adata, max_knn, dist_slot, key_added, njobs)


# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# -------------------------- ** NEIGHBOR GRAPH FUNC ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def neighbors_graph(adata,
                    n_neighbors=15,
                    knn_slot='knn'):
    """\
    A tool for rapidly computing a k-nearest neighbor (knn) graph (i.e.
    connectivities) that can then be used for clustering.

    graphs with acdc.pp.neighbors_graph for clustering.

    Parameters
    ----------
    adata
        An anndata object containing a distance object in adata.obsp.
    n_neighbors (default: 15)
        The number of nearest neighbors to use to build the connectivity graph.
        This number must be less than the total number of knn in the knn array
        stored in adata.uns[knn_slot].
    knn_slot (default: 101)
        The slot in adata.uns where the knn array is stored. One way of
        generating this object is with acdc.pp.neighbors_knn.

    Returns
    -------
    Adds fields to the input adata, such that it contains a knn graph stored in
    adata.obsp['connectivities'] along with metadata in adata.uns["neighbors"].
    """
    # returns if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
    return _neighbors_graph(adata, n_neighbors, knn_slot)



### ---------- IMPORT DEPENDENCIES ----------
import numpy as np
import pandas as pd
import anndata
import random
from .pp import corr_distance, neighbors_knn
from ._diffuse_correction_funcs import _correct_diffusion_with_leiden
from tqdm import tqdm
from math import log2

### ---------- EXPORT LIST ----------
# __all__ = ["subsample_anndata", "condense_anndata", "diffuse_subsample_labels"]

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ----------------------- ** CONDENSING ANNDATA FUNCS ** -----------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

# ----------------------------- ** HELPER FUNCS ** -----------------------------

def _representative_subsample_pca(pca_array, size, njobs = 1):
    dist_mat = corr_distance(pca_array)
    return representative_subsample(dist_mat, size, njobs)

def representative_subsample(dist_mat, size, njobs = 1):
    n_total_samples = dist_mat.shape[0]
    NN_array = neighbors_knn(dist_mat, n_total_samples, njobs)

    # If sample size is greater than half the number of samples, we take a
    # representative subsample of size (total # cells - sample size) and then
    # remove that representative subsample from all the possible samples. What's
    # left is our final representative subsample that we return.
    if size > int(n_total_samples/2):
        sample_size = n_total_samples - size
    else:
        sample_size = size

    rep_MC_sampling_df = more_equal_sampling_with_nn(NN_array, sample_size, seed = 0)
    first_MC_sample_representative_indices = rep_MC_sampling_df[rep_MC_sampling_df['sample_index'] != -1]['sample_index'].tolist()

    if size > int(n_total_samples/2):
        all_indices = list(range(0, n_total_samples))
        remaining_MC_indices = list(np.setdiff1d(all_indices, first_MC_sample_representative_indices))
        repr_sample = remaining_MC_indices
    else:
        repr_sample = first_MC_sample_representative_indices

    return repr_sample

def more_equal_sampling_with_nn(NN_array, size, seed = 0):
    # We select metacells in a representative way: after selecting each metacell, we eliminate its nearest
    # unselected neighbor from the sampling pool. This way, if you sample more from one subpopulation/cluster of
    # metacells, you are less likely to do so and therefore more likely to sample from the other subpopulations.
    # The reason we use this approach and not just random.sample is because if we have a small population,
    # random.sample may miss it, while using more_equal_sampling_with_nn will capture it.
    n_cells = NN_array.shape[0]
    if size > n_cells:
        raise ValueError("size of " + str(size) + " is larger than the number of samples in NN_array, " + str(n_cells))
    df = pd.DataFrame({'sample_index': [-1] * size, 'nn_index': [-1] * size})
    random.seed(seed)
    randomized_indices = random.sample(list(range(0, n_cells)), n_cells)
    # indices_in_sample = np.array(np.zeros(size), dtype = int)
    excluded_from_sampling = list()
    n_added_samps = 0 # n_collected_samples
    for i in range(0, n_cells):
        index_i = randomized_indices[i]
        if not index_i in excluded_from_sampling:
            # Add index_i as our newest sample
            df.loc[n_added_samps]["sample_index"] = index_i
            # Prevent index_i from being treated as a neighbor
            # of a future sample
            excluded_from_sampling.append(index_i)
            # Identify the NN of index_i that has not
            # yet been added to excluded_from_sampling
            index_i_NN_sorted = NN_array[index_i,:]
            NN_not_excluded = index_i_NN_sorted[~np.isin(index_i_NN_sorted, excluded_from_sampling)]
            first_NN_not_excluded = NN_not_excluded[0]
            df.loc[n_added_samps]["nn_index"] = first_NN_not_excluded
            # Prevent index_i from being treated as a sample or
            # a different sample's neighbor
            excluded_from_sampling.append(first_NN_not_excluded)
            # If we've collected enough samples return our results
            n_added_samps += 1
            if n_added_samps == size:
                break
    return(df)

def condense_in_half(pca_array, metacell_indices_df, seed = 0, njobs = 1):
    pca_array = pca_array.copy()
    n_samples = pca_array.shape[0]
    if n_samples % 2 == 1: # check if odd number of samples
        # Generate a random index for the row to remove
        random.seed(seed)
        random_index = random.randint(0, pca_array.shape[0] - 1)
        # Remove the random row, so we now have an even number of samples
        pca_array = np.delete(pca_array, random_index, axis=0)
        metacell_indices_df = pd.DataFrame(np.delete(metacell_indices_df.values, random_index, axis=0))
        n_samples -= 1

    dist_array = corr_distance(pca_array)

    # Slowest step
    NN_array = neighbors_knn(dist_array, n_samples, njobs)

    n_half_samples = int(n_samples/2)
    sampled_indices_df = more_equal_sampling_with_nn(NN_array, size = n_half_samples, seed = seed)

    # Extract sample_index and nn_index as NumPy arrays
    sample_indices = sampled_indices_df['sample_index'].to_numpy()
    nn_indices = sampled_indices_df['nn_index'].to_numpy()

    # Use NumPy to index rows in pca_array and calculate the mean
    pca_array = np.mean(pca_array[np.vstack((sample_indices, nn_indices))], axis=0)

    # Use NumPy to extract the corresponding rows from metacell_indices_df
    sample_rows = metacell_indices_df.iloc[sample_indices].to_numpy()
    nn_rows = metacell_indices_df.iloc[nn_indices].to_numpy()
    # Stack the rows side by side to create a new array with the
    # original indices in each row, so we can later create the metacells
    metacell_indices_df = pd.DataFrame(np.hstack((sample_rows, nn_rows)))

    return {"pca_array":pca_array, "metacell_indices_df":metacell_indices_df}

def get_mc_indices_by_condensing(pca_array, size = 1000, exact_size = False, seed = 0, verbose = True, njobs = 1): #mode = "exact"
    n_samples = pca_array.shape[0]
    metacell_indices_df = pd.DataFrame(list(range(0, n_samples)))
    # Calculate the number of iterations necessary
    # n_samples/2^n = size ==> n = log2(n_samples / size)
    from math import log2
    total_iters = int(log2(n_samples / size)) # we floor because we want > size
    if verbose: progress_bar = tqdm(total=total_iters, desc="Condensing")
    for _ in range(total_iters): #i.e. while n_samples > size
        condense_results = condense_in_half(pca_array, metacell_indices_df, seed, njobs)
        pca_array = condense_results["pca_array"]
        metacell_indices_df = condense_results["metacell_indices_df"]
        if verbose: progress_bar.update(1)

    if exact_size:
        repr_MC_sample_indices = _representative_subsample_pca(pca_array, size, njobs = 1)
        metacell_indices_df = metacell_indices_df.loc[repr_MC_sample_indices].reset_index(drop=True)

    if verbose: progress_bar.close()

    return metacell_indices_df

# ------------------------------ ** MAIN FUNCS ** ------------------------------

def subsample_anndata(adata, pca_array = None, size = 1000, exact_size = False, seed = 0, verbose = True, njobs = 1):
    if pca_array is None:
        pca_array = adata.obsm["X_pca"].copy()
    metacell_indices_df = get_mc_indices_by_condensing(pca_array, size, exact_size, seed, verbose, njobs)
    sample_indices = np.array(metacell_indices_df.loc[:,0])

    adata_subsample = adata[sample_indices,:].copy()
    adata_subsample.uns["adata_full"] = adata
    adata_subsample.var = adata.var
    adata_subsample.uns["sample_indices"] = sample_indices
    adata_subsample.uns["metacell_indices_df"] = metacell_indices_df
    return adata_subsample

# def condense_anndata(adata, pca_array = None, size = 1000, exact_size = False, seed = 0, verbose = True, recompute_signature = False):
#     if adata.raw is None:
#         raise ValueError("adata.raw is None. Raw counts required to condense anndata.")
#     if pca_array is None:
#         pca_array = adata.obsm["X_pca"].copy()
#     metacell_indices_df = get_mc_indices_by_condensing(pca_array, size, exact_size, seed, verbose)
#     counts = adata.raw.X.copy()
#     # Initialize a new array to store the result
#     metacells = np.zeros((metacell_indices_df.shape[0], counts.shape[1]))
#     # Loop through each row in sampled_indices_df and sum the corresponding rows in counts
#     for i, row_indices in enumerate(metacell_indices_df.values):
#         metacells[i, :] = np.sum(counts[row_indices, :], axis=0)
#     mc_data = anndata.AnnData(metacells, dtype=np.float64)
#     mc_data.var = adata.raw.var
#     mc_data.uns["adata_full"] = adata
#     mc_data.uns["metacell_indices_df"] = metacell_indices_df
#
#     if recompute_signature:
#             sc.pp.filter_cells(adata, min_genes=200)
#             sc.pp.filter_genes(adata, min_cells=3)
#             sc.pp.normalize_total(adata, target_sum=1e4)
#             sc.pp.log1p(adata)
#             adata = adata[:, adata.uns['adata_full'].var.index]
#             sc.pp.scale(adata, max_value=10)
#             sc.tl.pca(adata, svd_solver='arpack', random_state=0)
#
#     return mc_data
#


# def get_rep_subsample_via_condensed_metacells(pca_array, size = 1000, exact_size = False, seed = 0, verbose = True):
#     metacell_indices_df = get_mc_indices_by_condensing(pca_array, size, exact_size, seed, verbose)
#     return {"sample_indices":np.array(metacell_indices_df.loc[:,0]), "metacell_indices_df":metacell_indices_df}




# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ------------------------ ** DIFFUSING ANNDATA FUNCS ** -----------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

# ----------------------------- ** HELPER FUNCS ** -----------------------------

# Diffuse clusters via metacell construction
def diffuse_subsample_labels_by_metacell_diffusion(
    adata_subsample,
    adata_full,
    metacell_indices_df,
    key_added = "clusters",
    verbose = True
):
    adata_full.obs[key_added] = "-1"
    n_metacells = adata_subsample.shape[0]
    if verbose: pbar = tqdm(desc = "Diffusion", total = n_metacells, position=0, leave=True)
    for i in range(0, n_metacells):
        # Take the samples used to construct the metacell our subsample was in
        subcell_indices = metacell_indices_df.loc[i,:]
        # Get the subsample's labels and give it to its neighbors in the metacell
        adata_full.obs[key_added].iloc[subcell_indices] = adata_subsample.obs[key_added].iloc[i]
        if verbose: pbar.update(1)
    if verbose: pbar.close()
    return adata_full
# def diffuse_across_remaining_unlabeled_by_distance_diffusion(adata, key_added = "clusters"):
#     # For those samples that were lost during metacell construction
#     # diffuse using neighbor votes, weighted by distance
#     pca_array = adata.obsm["X_pca"].copy()
#     dist_array = compute_corr_distance(pca_array)
#     adata.obsm['dist_obj'] = compute_corr_distance(pca_array)
#
#     # Create new clusters DF so we don't affect the old clustering
#     # as we use it to update our labels
#     new_clusters_df = adata.obs[key_added].copy()
#     for index in tqdm(range(0,adata.shape[0]), desc="Correction"):#samples_missing_diffusion:
#         # Set up voting DataFrame
#         voting_df = pd.DataFrame({"distance":dist_array[index,:], "cluster":adata.obs[key_added].values})
#         voting_df = voting_df.drop(index)
#         voting_df = voting_df[voting_df['cluster'] != -1]
#         voting_df['voting_weight'] = (1/voting_df['distance'])**2
#
#         # Group voting by cluster
#         grouped = voting_df.groupby('cluster', observed=True)
#
#         # Calculate the sum of 'voting_weight' for each cluster
#         sum_voting_weight = grouped['voting_weight'].sum()
#
#         # Calculate the total number of samples in each cluster
#         total_samples = grouped.size()
#
#         # Divide the sum of 'voting_weight' by the total number of samples for each cluster
#         result_df = (sum_voting_weight / total_samples).reset_index()
#
#         # Rename the resulting column to 'voting_weight_per_sample'
#         result_df.rename(columns={0: 'voting_weight_per_sample'}, inplace=True)
#
#         # Get cluster with highest vote weighted by size
#         clust_with_highest_vote = result_df.sort_values(by='voting_weight_per_sample', ascending=False).iloc[0]['cluster']
#
#         # Assign this cluster to the sample
#         new_clusters_df.iloc[index] = clust_with_highest_vote
#
#     adata.obs[key_added] = new_clusters_df
#     return adata

# ------------------------------- ** MAIN FUNC ** ------------------------------

def __diffuse_subsample_labels(adata_subsample,
                               res,
                               knn,
                               dist_slot = None,
                               use_reduction = True,
                               reduction_slot="X_pca",
                               adata_full=None,
                               metacell_indices_df=None,
                               key_added='clusters',
                               knn_slot='knn',
                               verbose=True,
                               seed = 0,
                               njobs = 1):
    if adata_full is None:
        adata_full = adata_subsample.uns["adata_full"]
    if metacell_indices_df is None:
        metacell_indices_df = adata_subsample.uns["metacell_indices_df"]
    adata_full = diffuse_subsample_labels_by_metacell_diffusion(
        adata_subsample,
        adata_full,
        metacell_indices_df,
        key_added,
        verbose
    )
    if dist_slot is None:
        if verbose: print("Computing distance object...")
        dist_slot = "corr_dist"
        corr_distance(adata_full,
                      use_reduction,
                      reduction_slot,
                      key_added=dist_slot)

    adata_full = _correct_diffusion_with_leiden(
        adata_full,
        res,
        knn,
        dist_slot,
        key_added,
        knn_slot,
        verbose,
        seed,
        njobs
    )

    if "SA_results_dict" in adata_subsample.uns.keys():
        adata_full.uns["SA_results_dict"] = adata_subsample.uns["SA_results_dict"]
    if "GS_results_dict" in adata_subsample.uns.keys():
        adata_full.uns["GS_results_dict"] = adata_subsample.uns["GS_results_dict"]
    return adata_full



from ._pl import _GS_search_space, _SA_search_space

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** PLOTTING FUNCS ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def GS_search_space(adata, plot_type = "sil_mean"):
    """\
    Get a heatmap of the search space traversed by Grid Search (GS).

    Parameters
    ----------
    adata
        An anndata object that was previously given to GS
    plot_type (default: "sil_mean")
         A column name in adata.uns["GS_results_dict"]["search_df"].
         Among other, options include "sil_mean" and "n_clust".
    """
    return _GS_search_space(adata, plot_type)

def SA_search_space(adata, plot_type = "sil_mean", plot_density = True):
    # https://stackoverflow.com/questions/16834861/create-own-colormap-using-matplotlib-and-plot-color-scale
    """\
    Get a dot plot of the search space traversed by Simulated Annealing (SA).

    Parameters
    ----------
    adata
        An anndata object that was previously given to GS
    plot_type (default: "sil_mean")
         A column name in adata.uns["GS_results_dict"]["search_df"].
         Among other, options include "sil_mean" and "n_clust".
    plot_density (default: True)
        Whether to plot density on the dotplot to identify regions that were
        highly traversed by SA.
    """
    return _SA_search_space(adata, plot_type, plot_density)



### ---------- IMPORT DEPENDENCIES ----------
import numpy as np
import pandas as pd
import anndata
from .pp import corr_distance, neighbors_knn, neighbors_graph
from tqdm import tqdm
from math import log2
import random
import leidenalg
from sklearn.metrics import silhouette_samples
import igraph as ig


# from scanpy._utils import get_igraph_from_adjacency
def get_igraph_from_adjacency(adjacency, directed=None):
    """Get igraph graph from adjacency matrix."""
    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shape[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es["weight"] = weights
    except KeyError:
        pass
    if g.vcount() != adjacency.shape[0]:
        logg.warning(
            f"The constructed graph has only {g.vcount()} nodes. "
            "Your adjacency matrix contained redundant nodes."
        )
    return g

### ---------- EXPORT LIST ----------
__all__ = ["_correct_diffusion_with_leiden"]


# Compute silhouette scores for all samples
def __compute_sil_with_missing_values(adata_full, dist_slot = None, key_added = "clusters"):
    n_samples = adata_full.shape[0]
    cluster_labels = adata_full.obs[key_added].values.astype(int)
    if dist_slot not in adata_full.obsp.keys():
        corr_distance(adata_full)
    indices_nonmissing_vals = np.array(np.where(cluster_labels != -1))[0]
    nonmissing_samples_silhouette_values = silhouette_samples(
        X = np.array(adata_full.obsp[dist_slot])[indices_nonmissing_vals[:, None], indices_nonmissing_vals],
        labels = cluster_labels[indices_nonmissing_vals],
        metric="precomputed"
    )
    samples_silhouette_values = np.zeros(shape = (n_samples,))
    samples_silhouette_values[indices_nonmissing_vals] = nonmissing_samples_silhouette_values
    return samples_silhouette_values

def __get_cluster_labels_for_correction_and_fixed_membership(adata_full, dist_slot = None, key_added = "clusters"):
    samples_silhouette_values = __compute_sil_with_missing_values(adata_full, dist_slot, key_added)

    n_samples = adata_full.shape[0]
    cluster_labels = adata_full.obs[key_added].values.astype(int)

    # Fix membership of high confidence samples so they won't be altered during Leiden correction
    clusts_names = np.unique(cluster_labels[cluster_labels != -1])
    is_membership_fixed = np.zeros(shape = (n_samples,)).astype(bool)
    for clust_name in clusts_names:
        # Within each cluster get the silhouette threshold
        clust_sil = samples_silhouette_values[cluster_labels == clust_name]
        sil_threshold = np.percentile(clust_sil, 25)
        # Samples of those clusters above the threshold should have fixed membership
        is_membership_fixed[(samples_silhouette_values >= sil_threshold) & (cluster_labels == clust_name)] = True

    # Fix membership only for high SS
    indices_below_sil_threshold = ~is_membership_fixed
    n_samples_below_sil_threshold = np.count_nonzero(indices_below_sil_threshold)

    # For all samples that have low SS and no fixed membership, replace their cluster label with a unique integer
    cluster_labels_for_correction = np.array(cluster_labels).astype(int)
    cluster_labels_for_correction[indices_below_sil_threshold] = range(0, n_samples_below_sil_threshold)+np.max(cluster_labels)
    return cluster_labels_for_correction, is_membership_fixed

def __get_adjacency_graph(adata_full, knn, dist_slot = None, knn_slot = 'knn', njobs = 1):
    if not (knn_slot in adata_full.uns.keys()):
        knn_array = neighbors_knn(
            adata_full.obsp[dist_slot],
            max_knn=knn,
            njobs = njobs
        )
    elif not (adata_full.uns[knn_slot].shape[0] >= knn):
        knn_array = neighbors_knn(
            adata_full.obsp[dist_slot],
            max_knn=knn,
            njobs = njobs
        )
    else:
        knn_array = adata_full.uns[knn_slot]

    new_graph = neighbors_graph(knn_array, knn)
    g = get_igraph_from_adjacency(new_graph, directed=True)
    return g

def __get_partition_from_diffused_clusters(cluster_labels_for_correction, g, res, seed = 0):
    partition_type = leidenalg.RBConfigurationVertexPartition
    partition_weights = np.array(g.es['weight']).astype(np.float64)
    # Create the leidenalg partition,
    # but do n_iterations=0 so we just get the partition object

    partition = leidenalg.find_partition(
        graph = g,
        partition_type = partition_type,
        initial_membership=cluster_labels_for_correction.astype(int),#adata_reclust.obs['small_clusters_forClusts'].values.astype(int),
        weights=partition_weights,#None,
        n_iterations=0,#-1,#2,
        max_comm_size=0,
        seed=seed,
        resolution_parameter = res
    )

    return partition

def __get_corrected_clusters_from_leiden_optimization_of_parition(
    partition,
    is_membership_fixed,
    seed = 0
):
    # Get the Leiden algorithm optimizer
    opt = leidenalg.Optimiser()
    # https://github.com/vtraag/leidenalg/issues/89
    # If you then turn off the option to consider also empty communities for
    # moving nodes, the optimiser only considers existing communities, i.e.
    # it will never go beyond the number of communities already in our partition.
    opt.consider_empty_community = False
    opt.set_rng_seed(seed) # For consistent results
    opt.optimise_partition(partition, n_iterations=-1, is_membership_fixed=list(is_membership_fixed))

    corrected_clusters = np.array(partition.membership).astype(str)
    return corrected_clusters

def __merge_newly_created_cluster_into_original_clusters(adata_full, dist_slot, corrected_clusters, diffused_clusters):
    corrected_clusts_names = np.unique(corrected_clusters)
    old_clusts_names = np.unique(diffused_clusters)
    new_clusts_names = corrected_clusts_names[~np.isin(corrected_clusts_names.astype(str), old_clusts_names.astype(str))]
    n_new_clusts = len(new_clusts_names)

    if n_new_clusts > 0:
        for new_clust in new_clusts_names:
            max_total_sil = -1
            old_clust_for_merge = -1
            for old_clust in old_clusts_names:
                corrected_clusters_tmp = corrected_clusters.copy()
                # See what the silhouette is when merging new_clust into old_clust
                corrected_clusters_tmp[corrected_clusters == new_clust] = old_clust
                samples_silhouette_values_tmp = silhouette_samples(
                                X = adata_full.obsp[dist_slot],
                                labels = corrected_clusters_tmp,
                                metric="precomputed"
                            )
                total_sil = np.sum(samples_silhouette_values_tmp)
                if total_sil > max_total_sil:
                    old_clust_for_merge = old_clust
            corrected_clusters[corrected_clusters == new_clust] = old_clust_for_merge

    return corrected_clusters

def _correct_diffusion_with_leiden(adata_full,
                                   res,
                                   knn,
                                   dist_slot = None,
                                   key_added = 'clusters',
                                   knn_slot = 'knn',
                                   verbose = True,
                                   seed = 0,
                                   njobs = 1):
    if verbose: pbar = tqdm(desc = "Correction", total = 4, position=0, leave=True)

    diffused_clusters = adata_full.obs[key_added].values.astype(int)

    cluster_labels_for_correction, is_membership_fixed = \
        __get_cluster_labels_for_correction_and_fixed_membership(adata_full, dist_slot, key_added)
    if verbose: pbar.update(1)

    g = __get_adjacency_graph(adata_full, knn, dist_slot, knn_slot, njobs)
    if verbose: pbar.update(1)

    partition = __get_partition_from_diffused_clusters(cluster_labels_for_correction, g, res, seed)

    corrected_clusters = \
        __get_corrected_clusters_from_leiden_optimization_of_parition(partition, is_membership_fixed, seed)
    if verbose: pbar.update(1)

    corrected_clusters = \
        __merge_newly_created_cluster_into_original_clusters(adata_full, dist_slot, corrected_clusters, diffused_clusters)
    if verbose: pbar.update(1)
    if verbose: pbar.close()

    adata_full.obs[key_added] = corrected_clusters
    return adata_full



### ---------- IMPORT DEPENDENCIES ----------
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import seaborn as sns
import pandas as pd
import numpy as np

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# --------------------------- ** HELPER FUNCTIONS ** ---------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def _get_inferno_10():
    inferno_colors_10 = ["#000004FF",
                         "#1B0C42FF",
                         "#4B0C6BFF",
                         "#781C6DFF",
                         "#A52C60FF",
                         "#CF4446FF",
                         "#ED6925FF",
                         "#FB9A06FF",
                         "#F7D03CFF",
                         "#FCFFA4FF"]
    return(inferno_colors_10)
def _get_viridis_10():
    virids_colors_10 = ["#440154FF",
                        "#482878FF",
                        "#3E4A89FF",
                        "#31688EFF",
                        "#26828EFF",
                        "#1F9E89FF",
                        "#35B779FF",
                        "#6DCD59FF",
                        "#B4DE2CFF",
                        "#FDE725FF"]
    return(virids_colors_10)
def _get_YlOrRd_9():
    YlOrRd_9 = ["#FFFFCC", "#FFEDA0", "#FED976", "#FEB24C", "#FD8D3C", "#FC4E2A", "#E31A1C", "#BD0026", "#800026"]
    return(YlOrRd_9)

def _get_BWR_3():
    BWR_3 = ["blue", "white", "red"]
    return(BWR_3)

def _get_YlGnBu_9():
    YlGnBu_9 = ["#FFFFD9", "#EDF8B1", "#C7E9B4", "#7FCDBB", "#41B6C4", "#1D91C0", "#225EA8", "#253494", "#081D58"]
    return(YlGnBu_9)

def _get_RdPu_9():
    RdPu_9 = ["#FFF7F3", "#FDE0DD", "#FCC5C0", "#FA9FB5", "#F768A1", "#DD3497", "#AE017E", "#7A0177", "#49006A"]
    return(RdPu_9)
def _get_cmap_for_sa_search_plot(plot_type):
    if(plot_type == "sil_avg"):
        cmap = LinearSegmentedColormap.from_list("", _get_YlOrRd_9())#["red","violet","blue"])
    elif(plot_type == "iter"):
        cmap = LinearSegmentedColormap.from_list("", _get_YlGnBu_9())#["red","violet","blue"])
    elif(plot_type == "n_clust"):
        cmap = LinearSegmentedColormap.from_list("", _get_RdPu_9())
    else:
        cmap = LinearSegmentedColormap.from_list("", _get_inferno_10())
    return(cmap)
def _get_cbar_label_for_sa_search_plot(plot_type):
    if(plot_type == "sil_avg"):
        cbar_label = "Ave Sil Score"
    elif(plot_type == "iter"):
        cbar_label = "Iteration"
    elif(plot_type == "n_clust"):
        cbar_label = "Clusters"
    else:
        cbar_label = None
    return(cbar_label)
def _create_scatter_plot_for_sa_search_plot(ax, search_df, plot_type):
    cmap = _get_cmap_for_sa_search_plot(plot_type)
    ax.scatter(search_df["resolution"], search_df["knn"].astype(int), c=search_df[plot_type], cmap = cmap)
    ax.set_xlabel('Resolution')
    ax.set_ylabel('Nearest Neighbors')
def _create_cbar_for_sa_search_plot(ax, plot_type):
    cbar_label = _get_cbar_label_for_sa_search_plot(plot_type)
    PCM=ax.get_children()[0] #matplotlib.collections.PathCollection
    cbar = plt.colorbar(PCM, ax=ax)
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel(cbar_label, rotation=270)
def _create_countour_layer_for_sa_search_plot(ax, search_df):
    countour_layer = sns.kdeplot(x=search_df["resolution"],
                                 y=search_df["knn"],
                                 ax = ax,
                                 clip = [ax.get_xlim(),ax.get_ylim()])


# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** PLOTTING FUNCS ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def _GS_search_space(adata, plot_type = "sil_mean"):
    heatmap_table = pd.pivot_table(adata.uns["GS_results_dict"]["search_df"],
                                   values=plot_type,
                                   index=['knn'],
                                   columns=['resolution'],
                                   aggfunc=np.sum).astype(np.float64)
    # So that KNN are displayed as ints instead of numerics with a .0
    heatmap_table.index = heatmap_table.index.astype(int)

    if(plot_type == "sil_mean"):
        color_map = "YlOrRd"#"inferno"
    else: #plot_type = "n_clust"
        color_map = "YlGnBu"#"viridis"
    cbar_label = None
    if(plot_type == "sil_mean"):
        cbar_label = "Ave Sil Score"
    elif(plot_type == "n_clust"):
        cbar_label = "Clusters"

    fig = plt.figure()
    ax = sns.heatmap(heatmap_table,
                     cmap = color_map,
                     cbar_kws={'label': cbar_label},
                     linewidths=1,
                     linecolor='black')
    ax.invert_yaxis()
    # for _, spine in ax.spines.items():
        # spine.set_visible(True)
    plt.xlabel('Resolution')
    plt.ylabel('Nearest Neighbors')
    plt.close()
    return(fig)

def _SA_search_space(adata, plot_type = "sil_avg", plot_density = True):
    # https://stackoverflow.com/questions/16834861/create-own-colormap-using-matplotlib-and-plot-color-scale
    search_df = adata.uns["SA_results_dict"]["search_df"]
    fig, ax = plt.subplots()
    _create_scatter_plot_for_sa_search_plot(ax, search_df, plot_type)
    _create_cbar_for_sa_search_plot(ax, plot_type)
    if(plot_density == True):
        _create_countour_layer_for_sa_search_plot(ax, search_df)
    plt.close()
    return(fig)



from tqdm import tqdm
import warnings

import numpy as np
from scipy.optimize import OptimizeResult
from scipy.optimize import dual_annealing
from scipy.optimize import minimize, Bounds
from scipy.special import gammaln
from scipy._lib._util import check_random_state
from scipy.optimize._constraints import new_bounds_to_old
# Needed for this modified version of dual_annealing (with_progress_bar)
from scipy.optimize._dual_annealing import VisitingDistribution
from scipy.optimize._dual_annealing import EnergyState
from scipy.optimize._dual_annealing import StrategyChain
from scipy.optimize._dual_annealing import ObjectiveFunWrapper
from scipy.optimize._dual_annealing import LocalSearchWrapper

def dual_annealing_with_progress_bar(
    func,
    bounds,
    args=(),
    maxiter=1000,
    minimizer_kwargs=None,
    initial_temp=5230.,
    restart_temp_ratio=2.e-5,
    visit=2.62,
    accept=-5.0,
    maxfun=1e7,
    seed=None,
    no_local_search=False,
    callback=None,
    x0=None,
    local_search_options=None,
    show_progress_bar=True
):

    """
    Find the global minimum of a function using Dual Annealing.

    Parameters
    ----------
    func : callable
        The objective function to be minimized. Must be in the form
        ``f(x, *args)``, where ``x`` is the argument in the form of a 1-D array
        and ``args`` is a  tuple of any additional fixed parameters needed to
        completely specify the function.
    bounds : sequence or `Bounds`
        Bounds for variables. There are two ways to specify the bounds:

        1. Instance of `Bounds` class.
        2. Sequence of ``(min, max)`` pairs for each element in `x`.

    args : tuple, optional
        Any additional fixed parameters needed to completely specify the
        objective function.
    maxiter : int, optional
        The maximum number of global search iterations. Default value is 1000.
    minimizer_kwargs : dict, optional
        Extra keyword arguments to be passed to the local minimizer
        (`minimize`). Some important options could be:
        ``method`` for the minimizer method to use and ``args`` for
        objective function additional arguments.
    initial_temp : float, optional
        The initial temperature, use higher values to facilitates a wider
        search of the energy landscape, allowing dual_annealing to escape
        local minima that it is trapped in. Default value is 5230. Range is
        (0.01, 5.e4].
    restart_temp_ratio : float, optional
        During the annealing process, temperature is decreasing, when it
        reaches ``initial_temp * restart_temp_ratio``, the reannealing process
        is triggered. Default value of the ratio is 2e-5. Range is (0, 1).
    visit : float, optional
        Parameter for visiting distribution. Default value is 2.62. Higher
        values give the visiting distribution a heavier tail, this makes
        the algorithm jump to a more distant region. The value range is (1, 3].
    accept : float, optional
        Parameter for acceptance distribution. It is used to control the
        probability of acceptance. The lower the acceptance parameter, the
        smaller the probability of acceptance. Default value is -5.0 with
        a range (-1e4, -5].
    maxfun : int, optional
        Soft limit for the number of objective function calls. If the
        algorithm is in the middle of a local search, this number will be
        exceeded, the algorithm will stop just after the local search is
        done. Default value is 1e7.
    seed : {None, int, `numpy.random.Generator`,
            `numpy.random.RandomState`}, optional

        If `seed` is None (or `np.random`), the `numpy.random.RandomState`
        singleton is used.
        If `seed` is an int, a new ``RandomState`` instance is used,
        seeded with `seed`.
        If `seed` is already a ``Generator`` or ``RandomState`` instance then
        that instance is used.
        Specify `seed` for repeatable minimizations. The random numbers
        generated with this seed only affect the visiting distribution function
        and new coordinates generation.
    no_local_search : bool, optional
        If `no_local_search` is set to True, a traditional Generalized
        Simulated Annealing will be performed with no local search
        strategy applied.
    callback : callable, optional
        A callback function with signature ``callback(x, f, context)``,
        which will be called for all minima found.
        ``x`` and ``f`` are the coordinates and function value of the
        latest minimum found, and ``context`` has value in [0, 1, 2], with the
        following meaning:

            - 0: minimum detected in the annealing process.
            - 1: detection occurred in the local search process.
            - 2: detection done in the dual annealing process.

        If the callback implementation returns True, the algorithm will stop.
    x0 : ndarray, shape(n,), optional
        Coordinates of a single N-D starting point.
    local_search_options : dict, optional
        Backwards compatible flag for `minimizer_kwargs`, only one of these
        should be supplied.

        .. deprecated:: 1.8.0
            dual_annealing argument `local_search_options` is deprecated in
            favor of `minimizer_kwargs` and will be removed in SciPy 1.10.0.
    show_progress_bar : bool, optional
        If `show_progress_bar` is set to True, a progress bar will be displayed
        showing the current progress of the algorithm based on maxiter. This
        bar may exceed 100% under the condition that additional searches are
        needed by the algorithm.

    Returns
    -------
    res : OptimizeResult
        The optimization result represented as a `OptimizeResult` object.
        Important attributes are: ``x`` the solution array, ``fun`` the value
        of the function at the solution, and ``message`` which describes the
        cause of the termination.
        See `OptimizeResult` for a description of other attributes.

    Notes
    -----
    This function implements the Dual Annealing optimization. This stochastic
    approach derived from [3]_ combines the generalization of CSA (Classical
    Simulated Annealing) and FSA (Fast Simulated Annealing) [1]_ [2]_ coupled
    to a strategy for applying a local search on accepted locations [4]_.
    An alternative implementation of this same algorithm is described in [5]_
    and benchmarks are presented in [6]_. This approach introduces an advanced
    method to refine the solution found by the generalized annealing
    process. This algorithm uses a distorted Cauchy-Lorentz visiting
    distribution, with its shape controlled by the parameter :math:`q_{v}`

    .. math::

        g_{q_{v}}(\\Delta x(t)) \\propto \\frac{ \\
        \\left[T_{q_{v}}(t) \\right]^{-\\frac{D}{3-q_{v}}}}{ \\
        \\left[{1+(q_{v}-1)\\frac{(\\Delta x(t))^{2}} { \\
        \\left[T_{q_{v}}(t)\\right]^{\\frac{2}{3-q_{v}}}}}\\right]^{ \\
        \\frac{1}{q_{v}-1}+\\frac{D-1}{2}}}

    Where :math:`t` is the artificial time. This visiting distribution is used
    to generate a trial jump distance :math:`\\Delta x(t)` of variable
    :math:`x(t)` under artificial temperature :math:`T_{q_{v}}(t)`.

    From the starting point, after calling the visiting distribution
    function, the acceptance probability is computed as follows:

    .. math::

        p_{q_{a}} = \\min{\\{1,\\left[1-(1-q_{a}) \\beta \\Delta E \\right]^{ \\
        \\frac{1}{1-q_{a}}}\\}}

    Where :math:`q_{a}` is a acceptance parameter. For :math:`q_{a}<1`, zero
    acceptance probability is assigned to the cases where

    .. math::

        [1-(1-q_{a}) \\beta \\Delta E] < 0

    The artificial temperature :math:`T_{q_{v}}(t)` is decreased according to

    .. math::

        T_{q_{v}}(t) = T_{q_{v}}(1) \\frac{2^{q_{v}-1}-1}{\\left( \\
        1 + t\\right)^{q_{v}-1}-1}

    Where :math:`q_{v}` is the visiting parameter.

    .. versionadded:: 1.2.0

    References
    ----------
    .. [1] Tsallis C. Possible generalization of Boltzmann-Gibbs
        statistics. Journal of Statistical Physics, 52, 479-487 (1998).
    .. [2] Tsallis C, Stariolo DA. Generalized Simulated Annealing.
        Physica A, 233, 395-406 (1996).
    .. [3] Xiang Y, Sun DY, Fan W, Gong XG. Generalized Simulated
        Annealing Algorithm and Its Application to the Thomson Model.
        Physics Letters A, 233, 216-220 (1997).
    .. [4] Xiang Y, Gong XG. Efficiency of Generalized Simulated
        Annealing. Physical Review E, 62, 4473 (2000).
    .. [5] Xiang Y, Gubian S, Suomela B, Hoeng J. Generalized
        Simulated Annealing for Efficient Global Optimization: the GenSA
        Package for R. The R Journal, Volume 5/1 (2013).
    .. [6] Mullen, K. Continuous Global Optimization in R. Journal of
        Statistical Software, 60(6), 1 - 45, (2014).
        :doi:`10.18637/jss.v060.i06`

    Examples
    --------
    The following example is a 10-D problem, with many local minima.
    The function involved is called Rastrigin
    (https://en.wikipedia.org/wiki/Rastrigin_function)

    >>> from scipy.optimize import dual_annealing
    >>> func = lambda x: np.sum(x*x - 10*np.cos(2*np.pi*x)) + 10*np.size(x)
    >>> lw = [-5.12] * 10
    >>> up = [5.12] * 10
    >>> ret = dual_annealing(func, bounds=list(zip(lw, up)))
    >>> ret.x
    array([-4.26437714e-09, -3.91699361e-09, -1.86149218e-09, -3.97165720e-09,
           -6.29151648e-09, -6.53145322e-09, -3.93616815e-09, -6.55623025e-09,
           -6.05775280e-09, -5.00668935e-09]) # random
    >>> ret.fun
    0.000000

    """

    if isinstance(bounds, Bounds):
        bounds = new_bounds_to_old(bounds.lb, bounds.ub, len(bounds.lb))

    # noqa: E501
    if x0 is not None and not len(x0) == len(bounds):
        raise ValueError('Bounds size does not match x0')

    lu = list(zip(*bounds))
    lower = np.array(lu[0])
    upper = np.array(lu[1])
    # Check that restart temperature ratio is correct
    if restart_temp_ratio <= 0. or restart_temp_ratio >= 1.:
        raise ValueError('Restart temperature ratio has to be in range (0, 1)')
    # Checking bounds are valid
    if (np.any(np.isinf(lower)) or np.any(np.isinf(upper)) or np.any(
            np.isnan(lower)) or np.any(np.isnan(upper))):
        raise ValueError('Some bounds values are inf values or nan values')
    # Checking that bounds are consistent
    if not np.all(lower < upper):
        raise ValueError('Bounds are not consistent min < max')
    # Checking that bounds are the same length
    if not len(lower) == len(upper):
        raise ValueError('Bounds do not have the same dimensions')

    # Wrapper for the objective function
    func_wrapper = ObjectiveFunWrapper(func, maxfun, *args)
    # Wrapper for the minimizer
    if local_search_options and minimizer_kwargs:
        raise ValueError("dual_annealing only allows either 'minimizer_kwargs' (preferred) or "
                         "'local_search_options' (deprecated); not both!")
    if local_search_options is not None:
        warnings.warn("dual_annealing argument 'local_search_options' is "
                      "deprecated in favor of 'minimizer_kwargs' and will be "
                      "removed in SciPy 1.10.0.",
                      category=DeprecationWarning, stacklevel=2)
        minimizer_kwargs = local_search_options

    # minimizer_kwargs has to be a dict, not None
    minimizer_kwargs = minimizer_kwargs or {}

    minimizer_wrapper = LocalSearchWrapper(
        bounds, func_wrapper, **minimizer_kwargs)

    # Initialization of random Generator for reproducible runs if seed provided
    rand_state = check_random_state(seed)
    # Initialization of the energy state
    energy_state = EnergyState(lower, upper, callback)
    energy_state.reset(func_wrapper, rand_state, x0)
    # Minimum value of annealing temperature reached to perform
    # re-annealing
    temperature_restart = initial_temp * restart_temp_ratio
    # VisitingDistribution instance
    visit_dist = VisitingDistribution(lower, upper, visit, rand_state)
    # Strategy chain instance
    strategy_chain = StrategyChain(accept, visit_dist, func_wrapper,
                                   minimizer_wrapper, rand_state, energy_state)
    need_to_stop = False
    iteration = 0
    message = []
    # OptimizeResult object to be returned
    optimize_res = OptimizeResult()
    optimize_res.success = True
    optimize_res.status = 0

    t1 = np.exp((visit - 1) * np.log(2.0)) - 1.0
    # Run the search loop
    if show_progress_bar: pbar = tqdm(desc = "dual_annealing",
                                      total = maxiter,
                                      position=0,
                                      leave=True)
    while(not need_to_stop):
        for i in range(maxiter):
            # Compute temperature for this step
            s = float(i) + 2.0
            t2 = np.exp((visit - 1) * np.log(s)) - 1.0
            temperature = initial_temp * t1 / t2
            if iteration >= maxiter:
                message.append("Maximum number of iteration reached")
                need_to_stop = True
                break
            # Need a re-annealing process?
            if temperature < temperature_restart:
                energy_state.reset(func_wrapper, rand_state)
                break
            # starting strategy chain
            val = strategy_chain.run(i, temperature)
            if val is not None:
                message.append(val)
                need_to_stop = True
                optimize_res.success = False
                break
            # Possible local search at the end of the strategy chain
            if not no_local_search:
                val = strategy_chain.local_search()
                if val is not None:
                    message.append(val)
                    need_to_stop = True
                    optimize_res.success = False
                    break
            iteration += 1
            if show_progress_bar: pbar.update(1)

    if show_progress_bar: pbar.close()

    # Setting the OptimizeResult values
    optimize_res.x = energy_state.xbest
    optimize_res.fun = energy_state.ebest
    optimize_res.nit = iteration
    optimize_res.nfev = func_wrapper.nfev
    optimize_res.njev = func_wrapper.ngev
    optimize_res.nhev = func_wrapper.nhev
    optimize_res.message = message
    return optimize_res



from ._get_opt import _GS_clustering, _GS_metric_search_data, _GS_metric_value, _GS_params, _SA_clustering, _SA_metric_search_data, _SA_metric_value, _SA_params

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# -------------------------- ** OPTIMMIZATION FUNCS ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def SA_clustering(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    n_clusts = None,
    seed = 0,
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    key_added = "clusters",
    knn_slot = 'knn',
    verbose = True,
    njobs = 1
):
    """\
    Get the clustering using the parameters found by optimizing a particular
    metric using the results produced by the SA function. Note that this
    requires running the acdc.SA function first in order to produce
    adata.uns['SA_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing a distance object in adata.obsp.
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    dist_slot (default: None)
        Slot in adata.obsp where a pre-generated distance matrix computed across
        all cells is stored in adata for use in construction of NN. (Default =
        None, i.e. distance matrix will be automatically computed as a
        correlation distance and stored in "corr_dist").
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for clustering.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    clust_alg (default: "Leiden")
        Clustering algorithm. Choose among: "Leiden" (default) or  "Louvain".
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to compute the optimal clustering solution with
        this many clusters.
    seed (default: 0)
        Random seed to use.
    approx (default: {"run":False, "size":1000, "exact_size":True})
        A diciontary object containing three parameters to control subsampling and diffusion
            "run": True or False whether to use subsampling and diffusion. Default=True
            "size": the number of cells to use in the subsampling. Default=1000.
            "exact_size": whether to get the exact size "size" of subsampling (True) or
            be more inclusive during the representative subsampling (False, recommended).
    key_added (default: "clusters")
        Slot in obs to store the resulting clusters.
    knn_slot (default: "knn")
        Slot in uns that stores the KNN array used to compute a neighbors graph
        (i.e. adata.obs['connectivities']).
    verbose (default: True)
        Include additional output with True. Alternative = False.
    njobs (default: 1)
        Paralleization option that allows users to speed up runtime.

    Returns
    -------
    Adds fields to the input adata, such that it contains the clustering stored
    in adata.obs[key_added].
    """
    _SA_clustering(
        adata,
        opt_metric,
        opt_metric_dir,
        dist_slot,
        use_reduction,
        reduction_slot,
        clust_alg,
        n_clusts,
        seed,
        approx,
        key_added,
        knn_slot,
        verbose,
        njobs
    )

def SA_params(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal parameters for clustering found by optimizing a particular
    metric using the results produced by the SA function. Note that this
    requires running the acdc.SA function first in order to produce
    adata.uns['SA_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.SA function in
        adata.uns['SA_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    A dictionary with keys opt_res and opt_knn and the corresponding values that
    produce the requested clustering solution.
    """
    return _SA_params(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )

def SA_metric_value(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal value for a particular metric found when using parameters
    for clustering that optimize said metric. This will be identified in using
    the results produced by the SA function. Note that this therefore requires
    running acdc.GS first in order to produce adata.uns['SA_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.SA function in
        adata.uns['SA_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    The value of the metric when optimized.
    """
    return _SA_metric_value(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )

def SA_metric_search_data(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal parameters for clustering along with all their associated
    statistics. These will be found by optimizing a particular metric using the
    results produced by the SA function. Note that this requires running the
    acdc.SA function first in order to produce adata.uns['SA_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.SA function in
        adata.uns['SA_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    A pandas series containing the resolution and knn that produce the requested
    clustering solution along with all other metrics.
    """
    return _SA_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )

def GS_clustering(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    n_clusts = None,
    seed = 0,
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    key_added = "clusters",
    knn_slot = 'knn',
    verbose = True,
    njobs = 1
):
    """\
    Get the clustering using the parameters found by optimizing a particular
    metric using the results produced by the GS function. Note that this
    requires running the acdc.GS function first in order to produce
    adata.uns['GS_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing a distance object in adata.obsp.
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    dist_slot (default: None)
        Slot in adata.obsp where a pre-generated distance matrix computed across
        all cells is stored in adata for use in construction of NN. (Default =
        None, i.e. distance matrix will be automatically computed as a
        correlation distance and stored in "corr_dist").
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for clustering.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    clust_alg (default: "Leiden")
        Clustering algorithm. Choose among: "Leiden" (default) or  "Louvain".
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to compute the optimal clustering solution with
        this many clusters.
    seed (default: 0)
        Random seed to use.
    approx (default: {"run":False, "size":1000, "exact_size":True})
        A diciontary object containing three parameters to control subsampling and diffusion
            "run": True or False whether to use subsampling and diffusion. Default=True
            "size": the number of cells to use in the subsampling. Default=1000.
            "exact_size": whether to get the exact size "size" of subsampling (True) or
            be more inclusive during the representative subsampling (False, recommended).
    key_added (default: "clusters")
        Slot in obs to store the resulting clusters.
    knn_slot (default: "knn")
        Slot in uns that stores the KNN array used to compute a neighbors graph
        (i.e. adata.obs['connectivities']).
    verbose (default: True)
        Include additional output with True. Alternative = False.
    njobs (default: 1)
        Paralleization option that allows users to speed up runtime.

    Returns
    -------
    Adds fields to the input adata, such that it contains the clustering stored
    in adata.obs[key_added].
    """
    _GS_clustering(
        adata,
        opt_metric,
        opt_metric_dir,
        dist_slot,
        use_reduction,
        reduction_slot,
        clust_alg,
        n_clusts,
        seed,
        approx,
        key_added,
        knn_slot,
        verbose,
        njobs
    )

def GS_params(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal parameters for clustering found by optimizing a particular
    metric using the results produced by the GS function. Note that this
    requires running the acdc.GS function first in order to produce
    adata.uns['GS_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.GS function in
        adata.uns['GS_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    A dictionary with keys opt_res and opt_knn and the corresponding values that
    produce the requested clustering solution.
    """
    return _GS_params(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )

def GS_metric_value(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal value for a particular metric found when using parameters
    for clustering that optimize said metric. This will be identified in using
    the results produced by the GS function. Note that this therefore requires
    running acdc.GS first in order to produce adata.uns['GS_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.GS function in
        adata.uns['GS_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    The value of the metric when optimized.
    """
    return _GS_metric_value(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )

def GS_metric_search_data(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    """\
    Get the optimal parameters for clustering along with all their associated
    statistics. These will be found by optimizing a particular metric using the
    results produced by the GS function. Note that this requires running the
    acdc.GS function first in order to produce adata.uns['GS_results_dict'].

    Parameters
    ----------
    adata
        An anndata object containing the results of the acdc.SA function in
        adata.uns['GS_results_dict'].
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    n_clusts (default: None)
        If not None, restrict the search space to the number of clusters equal
        to n_clusts in order to retrieve the parameters for optimal clustering
        with this many clusters.

    Returns
    -------
    A pandas series containing the resolution and knn that produce the requested
    clustering solution along with all other metrics.
    """
    return _GS_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )



### ---------- IMPORT DEPENDENCIES ----------
import numpy as np
from sklearn.metrics import silhouette_score
from sklearn.metrics import silhouette_samples
from sklearn.metrics import calinski_harabasz_score
from sklearn.metrics import davies_bouldin_score
import scanpy as sc
import pandas as pd
from random import sample
from random import seed
import matplotlib.pyplot as plt
from tqdm import tqdm
import time
from datetime import datetime
from ._condense_diffuse_funcs import subsample_anndata
import pickle

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ----------------------------- ** SIL DF FUNCS ** -----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def getEmptySilDF(nrow = 0, metrics = ['sil_mean']):
    if isinstance(metrics, str): metrics = [metrics]
    sil_df = pd.DataFrame(
        columns=['iter','n_pcs','resolution','knn','n_clust','subsamp_iter','seed'] + metrics,
        index=np.arange(nrow)
    )
    return(sil_df)
def at_least_min(n, min):
    if n < min:
        return(min)
    else:
        return(n)
def update_sil_df(sil_df,
                  n_pcs,
                  a_res,
                  a_nn,
                  tot_clusters,
                  subsamp_iter,
                  metrics_i,
                  seed,
                  curr_iter,
                  update_method = "iloc"):

    row_dict = metrics_i
    row_dict['iter'] = curr_iter
    row_dict['n_pcs'] = n_pcs
    row_dict['resolution'] = a_res
    row_dict['knn'] = a_nn
    row_dict['n_clust'] = tot_clusters
    row_dict['subsamp_iter'] = subsamp_iter
    row_dict['seed'] = seed

    new_iter_results = np.vectorize(row_dict.get, otypes=[float])(sil_df.columns.values)

    if(update_method == "loc"):
        sil_df.loc[curr_iter] = new_iter_results
    elif(update_method == "concat"):
        new_iter_results = pd.DataFrame([new_iter_results], columns=sil_df.columns.values)
        # pd.DataFrame([[curr_iter,
        #                                   n_pcs,
        #                                   a_res,
        #                                   a_nn,
        #                                   tot_clusters,
        #                                   subsamp_iter,
        #                                   silhouette_avgs_i,
        #                                   seed]],
        #                            columns=['iter',
        #                                     'n_pcs',
        #                                     'resolution',
        #                                     'knn',
        #                                     'n_clust',
        #                                     'subsamp_iter',
        #                                     'sil_avg',
        #                                     'seed'])
        sil_df = pd.concat([sil_df, new_iter_results], ignore_index = True)
    return(sil_df)
def compute_silhouette_samples(
    dist_mat,
    adata,
    pct_cells,
    SS_weights,
    SS_exp_base,
    i,
    cluster_labels
):
    # In case Leiden only finds 1 cluster, set SS=0 so it won't be selected as optimal.
    sil = silhouette_samples(X = dist_mat,
                             labels = cluster_labels,
                             metric="precomputed").tolist()
    if(pct_cells < 100):
        seed(i)
        sil = np.array(sample(sil, int(pct_cells/100*len(cluster_labels))), dtype=float)
    else:
        sil = np.array(sil, dtype=float)
    if SS_weights == 'exp':
        neg_sil_indices = np.where(sil < 0)
        sil[neg_sil_indices] = -1 * (SS_exp_base ** np.abs(sil[neg_sil_indices]))
    return sil
def add_clustering_results_to_sil_df_using_subsampling(dist_mat,
                                                       adata,
                                                       i,
                                                       pct_cells,
                                                       sil_df,
                                                       n_pcs,
                                                       a_res,
                                                       a_nn,
                                                       SS_weights,
                                                       SS_exp_base,
                                                       curr_iter,
                                                       metrics = "sil_mean",
                                                       update_method="iloc",
                                                       key_added = "clusters"):
    cluster_labels = adata.obs[key_added]
    if(len(adata.obs[key_added].unique())==1):
        metrics_i = {key: np.NaN for key in metrics}
    else:
        metrics_i = {}
        if np.any(np.isin(["sil_mean", "sil_mean_median", "tot_sil_neg", "lowest_sil_clust", "max_sil_clust"], metrics)):
            silhouette_samples_i = compute_silhouette_samples(dist_mat, adata, pct_cells, SS_weights, SS_exp_base, i, cluster_labels)
        if "sil_mean" in metrics:
            metrics_i["sil_mean"] = np.mean(silhouette_samples_i)
        if "sil_mean_median" in metrics:
            # Compute the silhouette scores for each cluster
            cluster_silhouette_scores = []
            for cluster in np.unique(cluster_labels):
                mask = (cluster_labels == cluster)
                cluster_silhouette = silhouette_samples_i[mask]
                cluster_silhouette_scores.append(cluster_silhouette)
            cluster_medians = [np.median(cluster) for cluster in cluster_silhouette_scores]
            metrics_i["sil_mean_median"] = np.mean(cluster_medians)  # Mean of all clusters' median silhouette scores
        if "tot_sil_neg" in metrics:
            metrics_i["tot_sil_neg"] = np.sum(silhouette_samples_i[silhouette_samples_i < 0])
        if "lowest_sil_clust" in metrics:
            metrics_i["lowest_sil_clust"] = np.min(silhouette_samples_i)
        if "max_sil_clust" in metrics:
            metrics_i["max_sil_clust"] = np.max(silhouette_samples_i)
        if "ch" in metrics:
            metrics_i["ch"] = calinski_harabasz_score(adata.X, cluster_labels)
        if "db" in metrics:
            metrics_i["db"] = davies_bouldin_score(adata.X, cluster_labels)

    tot_clusters = len(adata.obs[key_added].cat.categories)
    sil_df = update_sil_df(sil_df,
                           n_pcs,
                           a_res,
                           a_nn,
                           tot_clusters,
                           i,
                           metrics_i,
                           i,
                           curr_iter,
                           update_method)
    return(sil_df)

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ----------------------------- ** CLUSTER FUNCS ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def get_clust_func(clust_alg):
    if(str.lower(clust_alg) == "louvain"):
        clust_func = sc.tl.louvain
    else:
        clust_func = sc.tl.leiden
    return(clust_func)
def _cluster_adata(adata,
                   my_random_seed,
                   res,
                   clust_alg,
                   key_added="clusters"):
    clust_alg_func = get_clust_func(clust_alg)
    clust_alg_func(adata,
                   random_state=my_random_seed,
                   resolution=res,
                   key_added=key_added)
    return(adata)

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# -------------------------- ** CLUSTERING FUNC ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def get_approx_anndata(adata, approx, seed, verbose, njobs):
    approx["mode"] = "subsample"
    if approx['mode'] == "subsample":
        if verbose: print("Creating a representative subsample by condensing data...")
        adata = subsample_anndata(adata, None, approx['size'], approx['exact_size'], seed, verbose, njobs)
    elif approx['mode'] == "metacell_standard":
        raise ValueError("metacell_standard is not yet implemented.")
        # if verbose: print("Condensing data and running a standard Scanpy pipeline...")
        # adata = condense_anndata(adata, None, approx['size'], approx['exact_size'], seed, verbose, recompute_signature = True)
    else:
        raise ValueError('Unsupported approx["mode"]:' + str(approx['mode']))
    return adata



### ---------- IMPORT DEPENDENCIES ----------
import numpy as np
import pandas as pd
from ._pp import _corr_distance, _neighbors_knn, _neighbors_graph
from ._SA_GS_subfunctions import _cluster_adata
from ._condense_diffuse_funcs import __diffuse_subsample_labels

### ---------- EXPORT LIST ----------
__all__ = []


def _cluster_final_internal(adata,
                            res,
                            knn,
                            dist_slot = None,
                            clust_alg="Leiden",
                            seed=0,
                            approx={
                                "run": False,
                                "size": 1000,
                                "exact_size": False
                            },
                            key_added="clusters",
                            knn_slot='knn',
                            verbose=True,
                            njobs = 1):
    if approx["run"] is True: adata = get_approx_anndata(adata, approx, seed, verbose, njobs)

    if verbose is True: print("Computing neighbor graph with " + str(knn) + " neighbors...")
    if not (knn_slot in adata.uns.keys()):
        _neighbors_knn(adata, max_knn=knn, dist_slot = dist_slot, key_added = knn_slot, njobs = njobs)
    elif not (adata.uns[knn_slot].shape[1] >= knn):
        _neighbors_knn(adata, max_knn=knn, dist_slot = dist_slot, key_added = knn_slot, njobs = njobs)
    _neighbors_graph(adata, n_neighbors = knn, knn_slot = knn_slot)

    if verbose is True: print("Clustering with resolution " + str(res) + " using " + str(clust_alg) + "...")
    adata = _cluster_adata(adata,
                           seed,#my_random_seed,
                           res,
                           clust_alg,
                           key_added)

    if approx["run"] is True:
        if verbose is True: print("Diffusing clustering results...")
        adata = __diffuse_subsample_labels(
            adata,
            res,
            knn,
            dist_slot,
            use_reduction,
            reduction_slot,
            key_added = key_added,
            knn_slot = knn_slot,
            verbose = verbose,
            seed = seed,
            njobs = njobs)

    return adata

def _cluster_final(adata,
                  res,
                  knn,
                  dist_slot=None,
                  use_reduction=True,
                  reduction_slot="X_pca",
                  clust_alg="Leiden",
                  seed=0,
                  approx={
                      "run": False,
                      "size": 1000,
                      "exact_size": False
                  },
                  key_added="clusters",
                  knn_slot='knn',
                  verbose=True,
                  njobs = 1):
    if dist_slot is None:
        if verbose: print("Computing distance object...")
        dist_slot = "corr_dist"
        _corr_distance(adata,
                      use_reduction,
                      reduction_slot,
                      key_added=dist_slot)

    adata = _cluster_final_internal(adata,
                                    res,
                                    knn,
                                    dist_slot,
                                    clust_alg,
                                    seed,
                                    approx,
                                    key_added,
                                    knn_slot,
                                    verbose,
                                    njobs)
    return adata



### ---------- IMPORT DEPENDENCIES ----------
import numpy as np
import pandas as pd
from scipy.stats import rankdata
from scipy.sparse import csr_matrix
from joblib import Parallel, delayed

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** DISTANCE FUNCS ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# SA_GS_subfunctions.R
def _corr_distance_matrix(data):
    # Equivalent to the following in R: d = sqrt(1 - stats::corr(X))
    # Computing the correlation
    corr_matrix = np.corrcoef(data)
    # Calculating sqrt_one_minus_corr_matrix
    sqrt_one_minus_corr_matrix = np.sqrt(1 - corr_matrix)
    # Ensuring diagonal contains 0 values
    np.fill_diagonal(sqrt_one_minus_corr_matrix, 0)
    return(sqrt_one_minus_corr_matrix)

def __add_row_column_names_to_dist_mat(dist_mat, adata):
    # Turn into a dataframe with row and column names
    df_dist = pd.DataFrame(
        dist_mat,
        columns = adata.obs_names,
        index = adata.obs_names
    )
    return(df_dist)

def _corr_distance(adata,
                   use_reduction=True,
                   reduction_slot="X_pca",
                   key_added="corr_dist"):
    if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
        return _corr_distance_matrix(adata)

    if use_reduction == False:
        # use original features
        d = _corr_distance_matrix(adata.X)
    elif use_reduction == True:
        # use principal components
        X = adata.obsm[reduction_slot]
        d = _corr_distance_matrix(X)
    else:
        raise ValueError("reduction must be logical.")
    d = __add_row_column_names_to_dist_mat(d, adata)
    adata.obsp[key_added] = d

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ---------------------------- ** KNN ARRAY FUNC ** ----------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def __rank_column(column): return rankdata(column, method='ordinal')

def __get_top_n_indices(dist_array_ranked, knn, njobs = 1):
    # Sort each row of the input array and get the indices of the sorted elements
    if njobs == 1:
        sorted_indices = np.argsort(dist_array_ranked, axis=1)
    else:
        sorted_indices = np.array(Parallel(n_jobs=njobs)(
            delayed(np.argsort)(row) for row in dist_array_ranked
        ))

    # Slice the sorted indices to keep only the top knn indices for each row
    top_n_indices = sorted_indices[:, :knn]

    return top_n_indices

def _get_knn_array(dist_df, max_knn, njobs = 1):
    # For each sample, sort neighbors by distance.
    if njobs ==1:
        dist_array_ranked = np.apply_along_axis(
            __rank_column,
            axis=1,
            arr=np.array(dist_df)
        )
    else:
        dist_array_ranked = np.array(Parallel(n_jobs=njobs)(
            delayed(__rank_column)(column) for column in dist_df
        ))

    # Trim this ranking to MaxNN
    max_knn_sorted_indices_array = __get_top_n_indices(dist_array_ranked, max_knn, njobs)

    return max_knn_sorted_indices_array

# ------------------------------ ** HELPER FUNC ** -----------------------------
def _neighbors_knn(adata,
                   max_knn=101,
                   dist_slot="corr_dist",
                   key_added="knn",
                   njobs = 1):
    if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
        return _get_knn_array(dist_df = adata, max_knn = max_knn, njobs = njobs)

    # Get the distance DataFrame
    dist_df = adata.obsp[dist_slot]

    # Comptue the KNN array
    max_knn_sorted_indices_array = _get_knn_array(dist_df, max_knn, njobs)

    # Add this to the adata
    adata.uns[key_added] = max_knn_sorted_indices_array

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# -------------------------- ** NEIGHBOR GRAPH FUNC ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

# ------------------------------ ** HELPER FUNC ** -----------------------------

def __get_connectivity_graph(max_knn_sorted_indices_array, knn):
    num_rows = max_knn_sorted_indices_array.shape[0]

    # Create a square array A filled with zeros
    A = np.zeros((num_rows, num_rows), dtype=int)

    # Slice max_knn_sorted_indices_array to keep only the top knn columns
    B = max_knn_sorted_indices_array[:, :knn]

    # Create row indices for indexing A
    row_indices = np.arange(num_rows).reshape(-1, 1)

    # Broadcast row indices to match the shape of B
    row_indices = np.tile(row_indices, (1, knn))

    # Use advanced indexing to set the values in A to 1 based on indices from B
    A[row_indices, B] = 1

    return A

# ------------------------------- ** MAIN FUNC ** ------------------------------
def _neighbors_graph(adata,
                     n_neighbors=15,
                     knn_slot='knn'):
    """\
    A tool for rapidly computing a k-nearest neighbor (knn) graph (i.e.
    connectivities) that can then be used for clustering.

    graphs with acdc.pp.neighbors_graph for clustering.

    Parameters
    ----------
    adata
        An anndata object containing a distance object in adata.obsp.
    n_neighbors (default: 15)
        The number of nearest neighbors to use to build the connectivity graph.
        This number must be less than the total number of knn in the knn array
        stored in adata.uns[knn_slot].
    knn_slot (default: 101)
        The slot in adata.uns where the knn array is stored. One way of
        generating this object is with acdc.pp.neighbors_knn.

    Returns
    -------
    Adds fields to the input adata, such that it contains a knn graph stored in
    adata.obsp['connectivities'] along with metadata in adata.uns["neighbors"].
    """
    if isinstance(adata, np.ndarray) or isinstance(adata, pd.DataFrame):
        return __get_connectivity_graph(adata, n_neighbors)

    new_graph = __get_connectivity_graph(adata.uns[knn_slot], n_neighbors)
    adata.obsp['connectivities'] = csr_matrix(new_graph)
    # adata.uns["neighbors"] = {'connectivities':adata.obsp['connectivities']}
    adata.uns["neighbors"] = {
        'connectivities_key': 'connectivities',
        'params': {'n_neighbors': n_neighbors}
    }



### ---------- IMPORT DEPENDENCIES ----------
from ._SA_GS_subfunctions import *
from ._SA_GS_subfunctions import _cluster_adata
from ._condense_diffuse_funcs import __diffuse_subsample_labels
from ._tl import _cluster_final_internal
from .pp import corr_distance, neighbors_knn, neighbors_graph
from tqdm import tqdm
from datetime import datetime
from multiprocessing import cpu_count
from joblib import Parallel, delayed

### ---------- EXPORT LIST ----------
__all__ = ['GS']

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# --------------------------- ** GRIDSEARCH FUNCS ** ---------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def get_results_for_knn(
    a_nn,
    adata,
    NN_vector,
    res_vector,
    n_pcs,
    metrics,
    seed,#my_random_seed,
    clust_alg,
    SS_weights,
    SS_exp_base,
    key_added,
    n_subsamples,
    subsamples_pct_cells,
    dist_slot
):
    n_iters = len(res_vector)*n_subsamples
    sil_df = getEmptySilDF(n_iters, metrics)
    curr_iter = np.where(np.isin(NN_vector, a_nn))[0][0] * len(res_vector)
    sil_df.index = np.arange(n_iters) + curr_iter

    neighbors_graph(adata, n_neighbors = a_nn)
    for a_res in res_vector:
        adata = _cluster_adata(adata,
                               seed,#my_random_seed,
                               a_res,
                               clust_alg,
                               key_added)
        # run 100 times, change the seed
        silhouette_avgs = []
        for i in range(1,n_subsamples+1):
            sil_df = add_clustering_results_to_sil_df_using_subsampling(
                               adata.obsp[dist_slot],
                               adata,
                               i,
                               subsamples_pct_cells,
                               sil_df,
                               n_pcs,
                               a_res,
                               a_nn,
                               SS_weights,
                               SS_exp_base,
                               curr_iter,
                               metrics,
                               update_method="loc",
                               key_added = key_added
            )
            curr_iter = curr_iter + 1
    return sil_df

def get_gs_results_1core(
    adata,
    NN_vector,
    res_vector,
    n_pcs,
    metrics,
    seed,#my_random_seed,
    clust_alg,
    SS_weights,
    SS_exp_base,
    key_added,
    n_subsamples,
    subsamples_pct_cells,
    dist_slot,
    show_progress_bar,
    verbose
):
    n_iters = len(NN_vector)*len(res_vector)*n_subsamples
    sil_df = getEmptySilDF(n_iters, metrics)
    curr_iter = 0
    if verbose: print("Beginning GridSearch clustering...")

    if show_progress_bar: pbar = tqdm(desc = "GridSearch", total = n_iters, position=0, leave=True)
    for a_nn in NN_vector:
        neighbors_graph(adata, n_neighbors = a_nn)
        for a_res in res_vector:
            adata = _cluster_adata(adata,
                                   seed,#my_random_seed,
                                   a_res,
                                   clust_alg,
                                   key_added)
            # run 100 times, change the seed
            silhouette_avgs = []
            for i in range(1,n_subsamples+1):
                sil_df = add_clustering_results_to_sil_df_using_subsampling(
                                   adata.obsp[dist_slot],
                                   adata,
                                   i,
                                   subsamples_pct_cells,
                                   sil_df,
                                   n_pcs,
                                   a_res,
                                   a_nn,
                                   SS_weights,
                                   SS_exp_base,
                                   curr_iter,
                                   metrics,
                                   update_method="loc",
                                   key_added = key_added
                )
                if show_progress_bar: pbar.update(1)
                curr_iter = curr_iter + 1
    if show_progress_bar: pbar.close()
    return sil_df

def get_gs_results_multicore(
    adata,
    NN_vector,
    res_vector,
    n_pcs,
    metrics,
    seed,#my_random_seed,
    clust_alg,
    SS_weights,
    SS_exp_base,
    key_added,
    n_subsamples,
    subsamples_pct_cells,
    dist_slot,
    verbose,
    njobs
):
    if verbose: print("Beginning GridSearch clustering...")
    sil_df = Parallel(njobs)(
        delayed(get_results_for_knn)(
            a_nn,
            adata,
            NN_vector,
            res_vector,
            n_pcs,
            metrics,
            seed,#my_random_seed,
            clust_alg,
            SS_weights,
            SS_exp_base,
            key_added,
            n_subsamples,
            subsamples_pct_cells,
            dist_slot
        ) for a_nn in NN_vector
    )
    sil_df = pd.concat(sil_df)
    return sil_df

def get_gs_results(
    adata,
    res_vector=np.arange(0.1, 2, 0.2),
    NN_vector=np.arange(11, 102, 10),
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    metrics="sil_mean",
    SS_weights="unitary",
    SS_exp_base=2.718282,
    verbose=True,
    show_progress_bar=True,
    clust_alg="Leiden",
    n_subsamples=1,
    subsamples_pct_cells=100,
    seed = 0,
    key_added = "clusters",
    njobs = 1
):
    # Consistency so that metrics is a list
    if isinstance(metrics, str): metrics = [metrics]

    if dist_slot is None:
        if verbose: print("Computing distance object...")
        dist_slot = "corr_dist"
        corr_distance(adata,
                      use_reduction,
                      reduction_slot,
                      key_added=dist_slot)

    if use_reduction == True:
        n_pcs = adata.obsm[reduction_slot].shape[1]
    else:
        n_pcs = None

    # ---------------- SUBSAMPLING HERE ----------------
    if verbose: print("Computing neighbors...")
    neighbors_knn(adata, max_knn=np.max(NN_vector), dist_slot=dist_slot, njobs = njobs)

    if njobs == 1:
        sil_df = get_gs_results_1core(
            adata,
            NN_vector,
            res_vector,
            n_pcs,
            metrics,
            seed,#my_random_seed,
            clust_alg,
            SS_weights,
            SS_exp_base,
            key_added,
            n_subsamples,
            subsamples_pct_cells,
            dist_slot,
            show_progress_bar,
            verbose
        )
    else:
        sil_df = get_gs_results_multicore(
            adata,
            NN_vector,
            res_vector,
            n_pcs,
            metrics,
            seed,#my_random_seed,
            clust_alg,
            SS_weights,
            SS_exp_base,
            key_added,
            n_subsamples,
            subsamples_pct_cells,
            dist_slot,
            verbose,
            njobs
        )

    sil_df["resolution"] = np.around(sil_df["resolution"].astype(np.double),3)#prevent 0.3 being 0.300000000004

    run_params = {
        "res_vector": res_vector,
        "NN_vector": NN_vector,
        "use_reduction": use_reduction,
        "reduction_slot": reduction_slot,
        "SS_weights": SS_weights,
        "SS_exp_base": SS_exp_base,
        "clust_alg": clust_alg,
        "n_subsamples": n_subsamples,
        "subsamples_pct_cells": subsamples_pct_cells,
    }
    gs_results = {
        "search_df": sil_df,
        "run_params": run_params
    }
    adata.uns["GS_results_dict"] = gs_results
    return(adata)

def get_opt_res_knn_from_gs_results(
    gs_results,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df = gs_results["search_df"]
    if n_clusts is not None:
        search_df = search_df[search_df['n_clust'] == n_clusts]

    if(opt_metric_dir == "max"):
        max_opt_metric = np.nanmax(search_df[opt_metric])
        search_df_opt_row = search_df[search_df[opt_metric] >= max_opt_metric].iloc[0]
    elif(opt_metric_dir == "min"):
        min_opt_metric = np.nanmin(search_df[opt_metric])
        search_df_opt_row = search_df[search_df[opt_metric] <= min_opt_metric].iloc[0]
    else:
        ValueError('Unsupported opt_metric_dir:' + str(opt_metric_dir) +
                    '\n\t opt_metric_dir must be "max" or "min".')

    opt_res = search_df_opt_row["resolution"]
    opt_knn = int(search_df_opt_row["knn"])
    opt_params = {"opt_res": opt_res, "opt_knn": opt_knn}
    return(opt_params)

# -------------------------- ** MAIN RUN FUNCTION ** ---------------------------
def GS(
    adata,
    res_vector = np.arange(0.1, 2, 0.2),
    NN_vector = np.arange(11, 102, 10),
    dist_slot = None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    metrics = "sil_mean", #["sil_mean", "sil_mean_median", "tot_sil_neg", "lowest_sil_clust","max_sil_clust"]
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    SS_weights = "unitary",
    SS_exp_base = 2.718282,
    n_subsamples = 1,
    subsamples_pct_cells = 100,
    seed = 0,
    key_added = "clusters",
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    verbose = True,
    show_progress_bar = True,
    njobs = 1
):
    """\
    A tool for the optimization-based unsupervised clustering of large-scale
    data. Grid Search (GS) allows for deterministic optimization of several
    variables—Nearest Neighbors and resolution–with several objective
    functions—e.g. Silhouette Score. An approximation method we call subsampling
    and diffusion is included to allow fast and accurate clustering of hundreds
    of thousands of cells.

    Parameters
    ----------
    adata
        An anndata object containing a gene expression signature in adata.X and
        gene expression counts in adata.raw.X.
    res_vector (default: np.arange(0.1, 2, 0.2))
         sequence of values of the resolution parameter.
    NN_vector (default: np.arange(11, 102, 10))
         sequence of values for the number of nearest neighbors.
    dist_slot (default: None)
        Slot in adata.obsp where a pre-generated distance matrix computed across
        all cells is stored in adata for use in construction of NN. (Default =
        None, i.e. distance matrix will be automatically computed as a
        correlation distance and stored in "corr_dist").
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for clustering.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    clust_alg (default: "Leiden")
        Clustering algorithm. Choose among: "Leiden" (default) or  "Louvain".
    metrics (default: "sil_mean")
        A metric or a list of metrics to be computed at each iteration of the
        GridSearch. Possible metrics to use include "sil_mean", "sil_mean_median",
        "tot_sil_neg", "lowest_sil_clust", "max_sil_clust", "ch" and "db".
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    SS_weights (default: "unitary")
        Negative silhouette scores can be given more weight by exponentiation ("exp").
        Otherwise, leave SS_weights as "unitary".
    SS_exp_base (default: 2.718282.)
        If SS_weights is set to "exp", then set the base for exponentiation.
    n_subsamples (default=1)
        Number of subsamples per bootstrap.
    subsamples_pct_cells (default: 100)
        Percentage of cells sample at each bootstrap iteration.
        i.e. when 100, 100%, all cells are used).
    seed (default: 0)
        Random seed to use.
    key_added (default: "clusters")
        Slot in obs to store the resulting clusters.
    approx (default: {"run":False, "size":1000, "exact_size":True})
        A diciontary object containing three parameters to control subsampling and diffusion
            "run": True or False whether to use subsampling and diffusion. Default=False
            "size": the number of cells to use in the subsampling. Default=1000.
            "exact_size": whether to get the exact size "size" of subsampling (True) or
            be more inclusive during the representative subsampling (False, recommended).
    verbose (default: True)
        Include additional output with True. Alternative = False.
    show_progress_bar (default: True)
        Show a progress bar to visualize the progress of the algorithm.
    njobs (default: 1)
        Paralleization option that allows users to speed up runtime.
    Returns
    -------
    A object of :class:~anndata.Anndata containing a clustering vector
    "clusters" in the .obs slot and a dictionary "GS_results_dict" with
    information on the run in the .uns slot.
    """
    if opt_metric not in metrics:
        raise ValueError("opt_metric (" + str(opt_metric) + ") is missing from metrics.")
    n_max_cores = cpu_count()
    if njobs > n_max_cores:
        raise ValueError('njobs (' + str(njobs) + ') is larger than the ' +
                         'number of CPU cores (' + str(n_max_cores) + ').')

    if approx is False: approx = {"run":False}
    if approx is True: approx = {"run":False}
    if "size" not in approx.keys(): approx["size"] = 1000
    if "exact_size" not in approx.keys(): approx["exact_size"] = False

    if approx["run"] is True: adata = get_approx_anndata(adata, approx, seed, verbose, njobs)

    adata = get_gs_results(
        adata,
        res_vector,
        NN_vector,
        dist_slot,
        use_reduction,
        reduction_slot,
        metrics,
        SS_weights,
        SS_exp_base,
        verbose,
        show_progress_bar,
        clust_alg,
        n_subsamples,
        subsamples_pct_cells,
        seed,
        key_added,
        njobs
    )

    opt_params = get_opt_res_knn_from_gs_results(
        adata.uns['GS_results_dict'],
        opt_metric,
        opt_metric_dir,
        n_clusts = None
    )
    adata = _cluster_final_internal(
        adata,
        opt_params["opt_res"],
        opt_params["opt_knn"],
        dist_slot,
        clust_alg,
        seed,
        approx = {"run":False},
        key_added = key_added,
        knn_slot = "knn",
        verbose = False,
        njobs = njobs
    )

    if approx["run"] is True:
        if verbose is True: print("Diffusing clustering results...")
        adata = __diffuse_subsample_labels(
            adata,
            res = opt_params["opt_res"],
            knn = opt_params["opt_knn"],
            dist_slot = dist_slot,
            use_reduction = use_reduction,
            reduction_slot = reduction_slot,
            key_added = key_added,
            knn_slot = 'knn',
            verbose = verbose,
            seed = seed,
            njobs = njobs
        )



import numpy as np
from .tl import cluster_final
from ._SA_GS_subfunctions import get_approx_anndata
from ._condense_diffuse_funcs import __diffuse_subsample_labels

### ---------- EXPORT LIST ----------
__all__ = []

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# --------------------------- ** HELPER FUNCTIONS ** ---------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def _results_metric_search_data(
    results,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df = results["search_df"]
    if n_clusts is not None:
        search_df = search_df[search_df['n_clust'] == n_clusts]

    if(opt_metric_dir == "max"):
        max_opt_metric = np.nanmax(search_df[opt_metric])
        search_df_opt_row = search_df[search_df[opt_metric] >= max_opt_metric].iloc[0]
    elif(opt_metric_dir == "min"):
        min_opt_metric = np.nanmin(search_df[opt_metric])
        search_df_opt_row = search_df[search_df[opt_metric] <= min_opt_metric].iloc[0]
    else:
        ValueError('Unsupported opt_metric_dir:' + str(opt_metric_dir) +
                    '\n\t opt_metric_dir must be "max" or "min".')
    return search_df_opt_row

def _generic_clustering(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    n_clusts = None,
    seed = 0,
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    key_added = "clusters",
    knn_slot = 'knn',
    verbose = True,
    njobs = 1,
    mode = "GS"
):
    if "size" not in approx.keys(): approx["size"] = 1000
    if "exact_size" not in approx.keys(): approx["exact_size"] = False

    if mode == "GS":
        opt_params = _GS_params(
            adata,
            opt_metric,
            opt_metric_dir,
            n_clusts
        )
    else:
        opt_params = _SA_params(
            adata,
            opt_metric,
            opt_metric_dir,
            n_clusts
        )

    if approx["run"] is True: adata = get_approx_anndata(adata, approx, seed, verbose, njobs)

    adata = cluster_final(adata,
                          res = opt_params["opt_res"],
                          knn = opt_params["opt_knn"],
                          dist_slot = dist_slot,
                          use_reduction = use_reduction,
                          reduction_slot = reduction_slot,
                          clust_alg = clust_alg,
                          seed = seed,
                          approx = {"run":False},
                          key_added = key_added,
                          knn_slot = knn_slot,
                          verbose = verbose,
                          njobs = njobs)
    if dist_slot is None: dist_slot = "corr_dist"

    if approx["run"] is True:
        if verbose is True: print("Diffusing clustering results...")
        adata = __diffuse_subsample_labels(
            adata,
            res = opt_params["opt_res"],
            knn = opt_params["opt_knn"],
            dist_slot = dist_slot,
            key_added = key_added,
            knn_slot = knn_slot,
            verbose = verbose,
            seed = seed,
            njobs = njobs
        )
    return adata


# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# -------------------------- ** OPTIMMIZATION FUNCS ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-

def _SA_clustering(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    n_clusts = None,
    seed = 0,
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    key_added = "clusters",
    knn_slot = 'knn',
    verbose = True,
    njobs = 1
):
    _generic_clustering(
        adata,
        opt_metric,
        opt_metric_dir,
        dist_slot,
        use_reduction,
        reduction_slot,
        clust_alg,
        n_clusts,
        seed,
        approx,
        key_added,
        knn_slot,
        verbose,
        njobs,
        mode = "SA"
    )

def _SA_params(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df_opt_row = _SA_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    opt_res = search_df_opt_row["resolution"]
    opt_knn = int(search_df_opt_row["knn"])
    opt_params = {"opt_res": opt_res, "opt_knn": opt_knn}
    return opt_params

def _SA_metric_value(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df_opt_row = _SA_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    return search_df_opt_row[opt_metric]

def _SA_metric_search_data(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    results = adata.uns['SA_results_dict']
    search_df_opt_row = _results_metric_search_data(
        results,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    return search_df_opt_row

def _GS_clustering(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg = "Leiden",
    n_clusts = None,
    seed = 0,
    approx = {
        "run":False,
        "size":1000,
        "exact_size":True
    },
    key_added = "clusters",
    knn_slot = 'knn',
    verbose = True,
    njobs = 1
):
    _generic_clustering(
        adata,
        opt_metric,
        opt_metric_dir,
        dist_slot,
        use_reduction,
        reduction_slot,
        clust_alg,
        n_clusts,
        seed,
        approx,
        key_added,
        knn_slot,
        verbose,
        njobs,
        mode = "GS"
    )

def _GS_params(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df_opt_row = _GS_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    opt_res = search_df_opt_row["resolution"]
    opt_knn = int(search_df_opt_row["knn"])
    opt_params = {"opt_res": opt_res, "opt_knn": opt_knn}
    return opt_params

def _GS_metric_value(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    search_df_opt_row = _GS_metric_search_data(
        adata,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    return search_df_opt_row[opt_metric]

def _GS_metric_search_data(
    adata,
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    n_clusts = None
):
    results = adata.uns['GS_results_dict']
    search_df_opt_row = _results_metric_search_data(
        results,
        opt_metric,
        opt_metric_dir,
        n_clusts
    )
    return search_df_opt_row



### ---------- IMPORT DEPENDENCIES ----------
from ._SA_GS_subfunctions import *
from ._SA_GS_subfunctions import _cluster_adata
from ._dual_annealing_with_progress_bar import *
from ._tl import _cluster_final_internal
from .pp import corr_distance, neighbors_knn, neighbors_graph
from ._condense_diffuse_funcs import __diffuse_subsample_labels
from sys import maxsize
from multiprocessing import cpu_count

### ---------- EXPORT LIST ----------
__all__ = ['SA']

# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
# ------------------------------------------------------------------------------
# ------------------------- ** DUAL ANNEALING FUNCS ** -------------------------
# ------------------------------------------------------------------------------
# @-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-@-
def get_sa_results(adata,
                   res_range = [0.1, 1.9],
                   NN_range = [11,101],
                   dist_slot = None,
                   use_reduction = True,
                   reduction_slot = "X_pca",
                   metrics = "sil_mean", #["sil_mean", "sil_mean_median", "tot_sil_neg", "lowest_sil_clust","max_sil_clust"]
                   opt_metric = "sil_mean",
                   opt_metric_dir = "max",
                   SS_weights = "unitary",
                   SS_exp_base = 2.718282,
                   verbose = True,
                   show_progress_bar = True,
                   clust_alg = "Leiden",
                   n_subsamples = 1,
                   subsamples_pct_cells=100,
                   maxiter = 20,#20,
                   initial_temp = 5230,
                   restart_temp_ratio = 2e-5,
                   visit = 2.62,
                   accept = -5.0,
                   maxfun = 1e7,
                   seed = 0,
                   key_added = "clusters",
                   njobs = 1):
    # par_init = NULL,
    # control = NULL,
    # lq = 0,
    # rng_seeds = c(1234,0)):
    if dist_slot is None:
        if verbose: print("Computing distance object...")
        dist_slot = "corr_dist"
        corr_distance(adata,
                      use_reduction,
                      reduction_slot,
                      key_added=dist_slot)
    if use_reduction == True:
        n_pcs = adata.obsm[reduction_slot].shape[1]
    else:
        n_pcs = None

    if verbose: print("Computing neighbors...")
    neighbors_knn(adata, max_knn=np.max(NN_range), dist_slot=dist_slot, njobs = njobs)

    # In order to use global inside a nested function, we have to declare a
        # variable with a global keyword inside a nested function
    global sil_df
    global curr_iter
    sil_df = getEmptySilDF(nrow = 0, metrics = metrics)
    curr_iter = 1
    bounds = [res_range, NN_range]

    def objective(v, adata, n_subsamples, subsamples_pct_cells,
                  n_pcs, clust_alg, SS_weights, SS_exp_base,
                  metrics, opt_metric, opt_metric_dir, dist_slot):
        a_res, a_nn = v
        a_nn = int(np.floor(a_nn))
        global sil_df
        global curr_iter
        neighbors_graph(adata, n_neighbors = a_nn)
        adata = _cluster_adata(adata,
                               seed,#my_random_seed,
                               a_res,
                               clust_alg,
                               key_added)
        # run 100 times, change the seed
        silhouette_avgs = []
        for i in range(1,n_subsamples+1):
            sil_df = add_clustering_results_to_sil_df_using_subsampling(
                               adata.obsp[dist_slot],
                               adata,
                               i,
                               subsamples_pct_cells,
                               sil_df,
                               n_pcs,
                               a_res,
                               a_nn,
                               SS_weights,
                               SS_exp_base,
                               curr_iter,
                               metrics,
                               update_method = "concat",
                               key_added = key_added
            )
        curr_iter = curr_iter + 1
        opt_metric_i = np.mean(sil_df[opt_metric].tail(n_subsamples).values) #Get the mean silouette score for these subsamples

        if opt_metric_dir == "max":
            if np.isnan(opt_metric_i):
                return 0
            else:
                return(opt_metric_i*-1)
        else:
            if np.isnan(opt_metric_i):
                return maxsize
            else:
                return(opt_metric_i)

    if verbose: print("Beginning Simulated Annealing clustering...")
    # perform the dual annealing search
    # opt_result = dual_annealing(func = objective,
    opt_result = dual_annealing_with_progress_bar(func = objective,
                                bounds = bounds,
                                args = (adata,
                                        n_subsamples,
                                        subsamples_pct_cells,
                                        n_pcs,
                                        clust_alg,
                                        SS_weights,
                                        SS_exp_base,
                                        metrics,
                                        opt_metric,
                                        opt_metric_dir,
                                        dist_slot),
                                maxiter = maxiter,
                                initial_temp = initial_temp,
                                restart_temp_ratio = restart_temp_ratio,
                                visit = visit,
                                accept = accept,
                                maxfun = maxfun,
                                seed = seed,
                                show_progress_bar = show_progress_bar)
    run_params = {
        "res_range": res_range,
        "NN_range": NN_range,
        "use_reduction": use_reduction,
        "reduction_slot": reduction_slot,
        "SS_weights": SS_weights,
        "SS_exp_base": SS_exp_base,
        "clust_alg": clust_alg,
        "n_subsamples": n_subsamples,
        "subsamples_pct_cells": subsamples_pct_cells,
        "maxiter": maxiter,
        "initial_temp": initial_temp,
        "restart_temp_ratio": restart_temp_ratio,
        "visit": visit,
        "accept": accept,
        "maxfun": maxfun,
        "seed": seed
    }
    sa_results = {
        "search_df": sil_df,
        "opt_result": opt_result,
        "run_params": run_params
    }
    adata.uns["SA_results_dict"] = sa_results
    return(adata)

# -------------------------- ** MAIN RUN FUNCTION ** ---------------------------
def SA(
    adata,
    res_range=[0.1, 1.9],
    NN_range=[11, 101],
    dist_slot=None,
    use_reduction=True,
    reduction_slot="X_pca",
    clust_alg="Leiden",
    metrics = "sil_mean", #["sil_mean", "sil_mean_median", "tot_sil_neg", "lowest_sil_clust","max_sil_clust"]
    opt_metric = "sil_mean",
    opt_metric_dir = "max",
    SS_weights="unitary",
    SS_exp_base=2.718282,
    n_subsamples=1,
    subsamples_pct_cells=100,
    maxiter=20,#20,
    initial_temp=5230,
    restart_temp_ratio=2e-5,
    visit=2.62,
    accept=-5.0,
    maxfun=1e7,
    seed=0,
    key_added = "clusters",
    approx = {
        "run":False,
        # "mode":"subsample", #subsample, metacell_standard,
        "size":1000,
        "exact_size":True
    },
    verbose=True,
    show_progress_bar = True,
    njobs = 1
):
    """\
    A tool for the optimization-based unsupervised clustering of large-scale
    data. Simulated Annealing (SA) allows for stochastic optimization of several
    variables—Nearest Neighbors and resolution–with several objective
    functions—e.g. Silhouette Score. An approximation method we call subsampling
    and diffusion is included to allow fast and accurate clustering of hundreds
    of thousands of cells.

    Parameters
    ----------
    adata
        An anndata object containing a gene expression signature in adata.X and
        gene expression counts in adata.raw.X.
    res_range (default: [0.1, 1.9])
         edge values of the search space for the resolution parameter.
    NN_range (default: [11, 101])
        edge values of the search space for the nearest neighbors parameter.
    dist_slot (default: None)
        Slot in adata.obsp where a pre-generated distance matrix computed across
        all cells is stored in adata for use in construction of NN. (Default =
        None, i.e. distance matrix will be automatically computed as a
        correlation distance and stored in "corr_dist").
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for clustering.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    clust_alg (default: "Leiden")
        Clustering algorithm. Choose among: "Leiden" (default) or  "Louvain".
    metrics (default: "sil_mean")
        A metric or a list of metrics to be computed at each iteration of the
        GridSearch. Possible metrics to use include "sil_mean", "sil_mean_median",
        "tot_sil_neg", "lowest_sil_clust", "max_sil_clust", "ch" and "db".
    opt_metric (default: "sil_mean")
        A metric from metrics to use to optimize parameters for the clustering.
    opt_metric_dir (default: "max")
        Whether opt_metric is more optimal by maximizing ("max") or
        by minimizing ("min").
    SS_weights (default: "unitary")
        Negative silhouette scores can be given more weight by exponentiation ("exp").
        Otherwise, leave SS_weights as "unitary".
    SS_exp_base (default: 2.718282.)
        If SS_weights is set to "exp", then set the base for exponentiation.
    n_subsamples (default=1)
        Number of subsamples per bootstrap.
    subsamples_pct_cells (default: 100)
        Percentage of cells sample at each bootstrap iteration.
        i.e. when 100, 100%, all cells are used).
    maxiter : (default: 20)
        The maximum number of global search iterations. If None, value is 1000.
    minimizer_kwargs : dict, optional
        Extra keyword arguments to be passed to the local minimizer
        (minimize). Some important options could be:
        method for the minimizer method to use and args for
        objective function additional arguments.
    initial_temp : float, optional
        The initial temperature, use higher values to facilitates a wider
        search of the energy landscape, allowing dual_annealing to escape
        local minima that it is trapped in. Default value is 5230. Range is
        (0.01, 5.e4].
    restart_temp_ratio : float, optional
        During the annealing process, temperature is decreasing, when it
        reaches initial_temp * restart_temp_ratio, the reannealing process
        is triggered. Default value of the ratio is 2e-5. Range is (0, 1).
    visit : float, optional
        Parameter for visiting distribution. Default value is 2.62. Higher
        values give the visiting distribution a heavier tail, this makes
        the algorithm jump to a more distant region. The value range is (1, 3].
    accept : float, optional
        Parameter for acceptance distribution. It is used to control the
        probability of acceptance. The lower the acceptance parameter, the
        smaller the probability of acceptance. Default value is -5.0 with
        a range (-1e4, -5].
    maxfun : int, optional
        Soft limit for the number of objective function calls. If the
        algorithm is in the middle of a local search, this number will be
        exceeded, the algorithm will stop just after the local search is
        done. Default value is 1e7.
    seed (default: 0)
        Random seed to use.
    key_added (default: "clusters")
        Slot in obs to store the resulting clusters.
    approx (default: {"run":False, "size":1000, "exact_size":True})
        A diciontary object containing three parameters to control subsampling and diffusion
            "run": True or False whether to use subsampling and diffusion. Default=False
            "size": the number of cells to use in the subsampling. Default=1000.
            "exact_size": whether to get the exact size "size" of subsampling (True) or
            be more inclusive during the representative subsampling (False, recommended).
    verbose (default: True)
        Include additional output with True. Alternative = False.
    show_progress_bar (default: True)
        Show a progress bar to visualize the progress of the algorithm.
    njobs (default: 1)
        Paralleization option that allows users to speed up runtime.
    Returns
    -------
    A object of :class:~anndata.Anndata containing a clustering vector
    "clusters" in the .obs slot and a dictionary "GS_results_dict" with
    information on the run in the .uns slot.
    """
    if opt_metric not in metrics:
        raise ValueError("opt_metric (" + str(opt_metric) + ") is missing from metrics.")
    n_max_cores = cpu_count()
    if njobs > n_max_cores:
        raise ValueError('njobs (' + str(njobs) + ') is larger than the ' +
                         'number of CPU cores (' + str(n_max_cores) + ').')

    if approx is False: approx = {"run":False}
    if approx is True: approx = {"run":False}
    if "size" not in approx.keys(): approx["size"] = 1000
    if "exact_size" not in approx.keys(): approx["exact_size"] = False

    if approx["run"] is True: adata = get_approx_anndata(adata, approx, seed, verbose, njobs)

    adata = get_sa_results(adata,
                           res_range,
                           NN_range,
                           dist_slot,
                           use_reduction,
                           reduction_slot,
                           metrics,
                           opt_metric,
                           opt_metric_dir,
                           SS_weights,
                           SS_exp_base,
                           verbose,
                           show_progress_bar,
                           clust_alg,
                           n_subsamples,
                           subsamples_pct_cells,
                           maxiter,
                           initial_temp,
                           restart_temp_ratio,
                           visit,
                           accept,
                           maxfun,
                           seed,
                           key_added,
                           njobs)
    sa_results = adata.uns["SA_results_dict"]

    opt_res = sa_results["opt_result"].x[0]
    opt_knn = int(np.floor(sa_results["opt_result"].x[1]))
    adata = _cluster_final_internal(
        adata,
        opt_res,
        opt_knn,
        dist_slot,
        clust_alg,
        seed,
        approx = {"run":False},
        key_added = key_added,
        knn_slot = "knn",
        verbose = False,
        njobs = njobs
    )

    if approx["run"] is True:
        if verbose is True: print("Diffusing clustering results...")
        adata = __diffuse_subsample_labels(
            adata,
            res = opt_res,
            knn = opt_knn,
            dist_slot = dist_slot,
            use_reduction = use_reduction,
            reduction_slot = reduction_slot,
            key_added = key_added,
            knn_slot = 'knn',
            verbose = verbose,
            seed = seed,
            njobs = njobs
        )

    return(adata)



import setuptools

with open("README.md", "r") as fh:
    description = fh.read()

setuptools.setup(
    name="acdc_py",
    version="1.0.0",
    author="Alexander L.E. Wang",
    author_email="aw3436@cumc.columbia.edu",
    packages=["acdc_py"],
    description="A package to quickly identify unbiased graph-based clusterings via parameter optimization in Python",
    long_description="This Scanpy-compatible package provides the ability to iterative compute clusterings using graph-based on combinations of resolution and KNN. At each point in this search space, a metric, such as silhouette score, is computed. The solution from this search space that maximizes the metric results in an optimal unbiased clustering solution. This process is sped up by our subsampling and diffusion feature that provides a near-identical solution in a much shorter time.",
    long_description_content_type="text/markdown",
    url="https://github.com/alexanderlewis99/acdc_py",
    license='MIT',
    python_requires='>=3.8',
    install_requires=[
        "scipy",
        "tqdm",
        "scanpy",
        "anndata",
        "pandas",
        "numpy",
        "joblib",
        "leidenalg",
        "scikit-learn",
        "matplotlib",
        "seaborn",
        "datetime"
    ]
)



from .SA import *
from .GS import *
from .pl import *
from .tl import *
from .pp import *
from .get_opt import *
# from ._condense_diffuse_funcs import *

# Want to appear as module - no direct functions
# made __all__ = [] in each of these .py files



from ._tl import _cluster_final

### ---------- EXPORT LIST ----------
__all__ = []

def cluster_final(adata,
                  res,
                  knn,
                  dist_slot=None,
                  use_reduction=True,
                  reduction_slot="X_pca",
                  clust_alg="Leiden",
                  seed=0,
                  approx={
                      "run": False,
                      "size": 1000,
                      "exact_size": False
                  },
                  key_added="clusters",
                  knn_slot='knn',
                  verbose=True,
                  njobs = 1):
    """\
    A tool for replicating the final optimization-based unsupervised clustering
    of large-scale data performed by the Grid Search (GS) or Simulated Annealing
    (SA) functions. This includes replicating the approximation method we call
    subsampling and diffusion that enables fast and accurate clustering of
    hundreds of thousands of cells.

    Parameters
    ----------
    adata
        An anndata object containing a gene expression signature in adata.X and
        gene expression counts in adata.raw.X.
    res
         sequence of values of the resolution parameter.
    knn
         sequence of values for the number of nearest neighbors.
    dist_slot (default: None)
        Slot in adata.obsp where a pre-generated distance matrix computed across
        all cells is stored in adata for use in construction of NN. (Default =
        None, i.e. distance matrix will be automatically computed as a
        correlation distance and stored in "corr_dist").
    use_reduction (default: True)
        Whether to use a reduction (True) (highly recommended - accurate & much faster)
        or to use the direct matrix (False) for clustering.
    reduction_slot (default: "X_pca")
        If reduction is TRUE, then specify which slot for the reduction to use.
    clust_alg (default: "Leiden")
        Clustering algorithm. Choose among: "Leiden" (default) or  "Louvain".
    seed (default: 0)
        Random seed to use.
    key_added (default: "clusters")
        Slot in obs to store the resulting clusters.
    knn_slot (default: "knn")
        Slot in uns that stores the KNN array used to compute a neighbors graph
        (i.e. adata.obs['connectivities']).
    approx (default: {"run":False, "size":1000, "exact_size":False})
        A diciontary object containing three parameters to control subsampling and diffusion
            "run": True or False whether to use subsampling and diffusion. Default=False
            "size": the number of cells to use in the subsampling. Default=1000.
            "exact_size": whether to get the exact size "size" of subsampling (True) or
            be more inclusive during the representative subsampling (False, recommended).
    verbose (default: True)
        Include additional output with True. Alternative = False.

    Returns
    -------
    A object of :class:~anndata.Anndata containing a clustering vector
    "clusters" in the .obs slot.
    """
    return _cluster_final(
      adata,
      res,
      knn,
      dist_slot,
      use_reduction,
      reduction_slot,
      clust_alg,
      seed,
      approx,
      key_added,
      knn_slot,
      verbose,
      njobs
    )
